import json
import os
from dataclasses import dataclass, field
from logging import getLogger
from pathlib import Path
from typing import Any, Dict


def default_log_config() -> Dict[str, Any]:
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {"format": "{asctime} {levelname} {message}", "style": "{"},
            "access": {
                "format": "{asctime} ACCESS {host}:{port} {request_id} {method} "
                "HTTP/{http_version} {path} {status} {message}",
                "style": "{",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "standard",
                "stream": "ext://sys.stdout",
            },
            "access": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "formatter": "access",
                "stream": "ext://sys.stdout",
            },
            "exception": {
                "class": "logging.StreamHandler",
                "level": "ERROR",
                "formatter": "standard",
                "stream": "ext://sys.stderr",
            },
        },
        "loggers": {
            "qt_server": {"handlers": ["console"], "level": "DEBUG"},
            "qt_child": {"handlers": ["console"], "level": "DEBUG"},
            "qt_access": {"handlers": ["access"], "level": "INFO"},
            "qt_exception": {"handlers": ["exception"], "level": "ERROR"},
        },
    }


@dataclass
class Config:
    HOST: str = "localhost"
    PORT: int = 8000
    ADMIN_HOST: str = "localhost"
    ADMIN_PORT: int = 1986
    CHECK_PROCESS_INTERVAL: int = 1
    SELECT_SLEEP_TIME: float = 0.025
    RECV_TIMEOUT: float = 0.001
    RECV_BYTES: int = 65536
    MAX_PROCESSES: int = 500
    GATHER_PROC_STATS: bool = False
    REQUEST_TIMEOUT: float = 5
    APPS: Dict[str, str] = field(default_factory=dict)
    LOGS: Dict[str, Any] = field(default_factory=default_log_config)


def config_init() -> Config:
    logger = getLogger("Qactuar")
    env_var_name = "QACTUAR_CONFIG"
    path_str = os.environ.get(env_var_name)
    config_path = Path(path_str or "")
    if not config_path.is_file():
        logger.warning(
            f"Config file path is not valid or not set, loading default values. To use "
            f"a config file set up an environment variable called '{env_var_name}' and "
            f"set the value to a JSON config file path."
        )
        return Config()
    return Config(**json.loads(config_path.open().read()))



import sys
from importlib import import_module

import qactuar


def main() -> None:
    if len(sys.argv) < 2:
        qactuar.make_server()
        sys.exit(0)
    app_path = sys.argv[1]
    module_str, app_str = app_path.split(":")
    module = import_module(module_str)
    app = getattr(module, app_str)
    qactuar.make_server(app=app)


if __name__ == "__main__":
    main()



from typing import Any

from tornado.httputil import HTTPServerRequest
from tornado.web import Application, RequestHandler

from qactuar.util import TornadoWrapper


# noinspection PyAbstractClass
class HelloWorld(RequestHandler):
    def __init__(
        self, application: Application, request: HTTPServerRequest, **kwargs: Any
    ):
        super().__init__(application, request, **kwargs)

    async def get(self) -> None:
        self.write("Hello, WORLD!")


app = TornadoWrapper(HelloWorld)



import argparse
import threading
from time import time

import requests

RESPONSES = []
ELAPSEDES = []
THREADS = []


def connect(num, host, port, path):
    response = requests.post(
        f"http://{host}:{port}{path}", data={"thing": LOREM_IPSUM, "num": num}
    )
    elapsed = response.elapsed
    ELAPSEDES.append(elapsed.microseconds)
    response = response.text
    # RESPONSES.append(response)
    print(f"{response} {elapsed}")


def start(max_clients, max_conns, host, port, path):
    start_time = time()
    for client_num in range(max_clients):
        for connection_num in range(max_conns):
            t = threading.Thread(target=connect, args=(client_num, host, port, path))
            t.daemon = True
            THREADS.append(t)
    for thread in THREADS:
        thread.start()
    for thread in THREADS:
        thread.join()
    for response in RESPONSES:
        print(response)
    end_time = time()
    total_time = end_time - start_time
    print(f"total time: {total_time}, avg: {(sum(ELAPSEDES)/len(ELAPSEDES))/1000} ms")


LOREM_IPSUM = """Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nulla eu 
commodo justo. Interdum et malesuada fames ac ante ipsum primis in faucibus. Aliquam 
erat volutpat. Aliquam risus eros, mattis et volutpat eget, gravida vitae justo. Nunc 
risus leo, tincidunt in egestas vitae, dapibus sed enim. Interdum et malesuada fames ac 
ante ipsum primis in faucibus. Sed posuere libero vitae tortor mollis interdum sed quis 
risus. Phasellus sed turpis at augue luctus efficitur vitae vitae urna. Duis in lacus 
vitae tortor efficitur elementum a sed mauris. Morbi pharetra feugiat faucibus. Quisque 
risus dui, venenatis at placerat ut, dictum consequat lacus. Donec egestas hendrerit 
quam quis eleifend.

Nulla dictum blandit ligula, a interdum ipsum iaculis ut. Ut porta, orci sed ultricies 
tempor, felis metus sodales felis, at convallis urna nisi in arcu. Pellentesque quis 
condimentum dolor, in volutpat mauris. Donec sollicitudin tempus felis, non aliquam 
ante fringilla vel. Nulla suscipit rhoncus erat ac vehicula. Nullam lacinia 
sollicitudin suscipit. Nullam faucibus placerat sem ut pretium. Nunc at aliquet nulla. 
Duis vitae porta orci. Etiam consequat ipsum eu massa laoreet, et tincidunt lectus 
ultrices. Vivamus id iaculis libero, ac porttitor nunc. Integer faucibus nibh velit, 
ac bibendum ipsum lobortis id. Morbi ut consectetur sem.

Aliquam erat volutpat. Cras in vulputate nisl. Curabitur quis purus neque. Vivamus a 
faucibus arcu. Sed sit amet dignissim orci. Nulla at condimentum turpis, sed tristique 
tellus. Vestibulum dictum imperdiet nibh ut lacinia. Vestibulum scelerisque odio est, 
sed congue enim hendrerit quis. Quisque venenatis orci non venenatis pellentesque. 
Integer consectetur tristique lacus, at sagittis erat tincidunt eu. Nunc varius, lectus 
in consequat tincidunt, purus erat consectetur sem, ut aliquet tellus neque vitae 
purus. Sed ultricies massa id eleifend faucibus. Duis laoreet ac justo auctor porta. 
Nunc nec magna in sapien congue scelerisque. In velit dui, placerat ut quam nec, semper 
sagittis dui. Nam vel sem ultricies, iaculis justo mattis, ultricies felis.

Sed tempus vehicula rutrum. Morbi porta magna ut purus placerat vulputate. Phasellus 
efficitur dolor eget elit cursus congue. Quisque sagittis eget diam quis finibus. 
Praesent dignissim velit ac quam sagittis, ut rutrum lorem venenatis. Vivamus aliquet 
ex vitae ex ornare auctor. In hac habitasse platea dictumst.

Mauris nec est quis massa sollicitudin tempus eu et nisi. Nunc luctus lacus facilisis 
quam laoreet lacinia. Morbi sagittis ligula quis ligula ornare tempus vel vel neque. 
Vivamus porta scelerisque metus. Interdum et malesuada fames ac ante ipsum primis in 
faucibus. Morbi vitae blandit erat. Nullam nec iaculis dolor."""


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Test client for Qactuar.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--max-conns",
        type=int,
        default=1,
        help="Maximum number of connections per client.",
    )
    parser.add_argument(
        "--max-clients", type=int, default=1, help="Maximum number of clients."
    )
    parser.add_argument(
        "--host", type=str, default="localhost", help="Host to connect to",
    )
    parser.add_argument(
        "--port", type=int, default=8000, help="Port to connect to",
    )
    parser.add_argument(
        "--path", type=str, default="/tonberry/what", help="URL path to connect access",
    )
    args = parser.parse_args()
    start(args.max_clients, args.max_conns, args.host, args.port, args.path)



async def app(scope, receice, send):
    await send({"type": "http_response.start", "status": "200", "headers": []})
    await send({"type": "http.response.body", "body": b"Hello", "more_body": False})



from typing import TYPE_CHECKING, Optional

from qactuar.child_process import ChildProcess
from qactuar.models import Message, Scope

if TYPE_CHECKING:
    from qactuar import QactuarServer

ASGI_VERSION = {"version": "2.0", "spec_version": "2.0"}


class Handler:
    def __init__(self, server: "QactuarServer"):
        self.server = server
        self.child: Optional[ChildProcess] = None

    def set_child(self, child: ChildProcess) -> None:
        self.child = child


class HTTPHandler(Handler):
    def create_scope(self) -> Scope:
        # TODO: Pseudo headers (present in HTTP/2 and HTTP/3) must be removed; if
        #  :authority is present its value must be added to the start of the iterable
        #  with host as the header name or replace any existing host header already
        #  present.
        if self.child:
            return {
                "type": "http",
                "asgi": ASGI_VERSION,
                "http_version": self.child.request_data.request_version_num,
                "method": self.child.request_data.method,
                "scheme": self.server.scheme,
                "path": self.child.request_data.path,
                "raw_path": self.child.request_data.raw_path,
                "query_string": self.child.request_data.query_string,
                "root_path": "",
                "headers": self.child.request_data.raw_headers,
                "client": self.server.client_info,
                "server": (self.server.server_name, self.server.server_port),
            }
        else:
            raise AttributeError

    async def receive(self) -> Message:
        # TODO: support streaming from client
        if self.child:
            return {
                "type": "http.request",
                "body": self.child.request_data.body,
                "more_body": False,
            }
        else:
            raise AttributeError

    async def send(self, data: Message) -> None:
        if self.child:
            if data["type"] == "http.response.start":
                self.child.response.status = str(data["status"]).encode("utf-8")
                self.child.response.headers += data["headers"]
            if data["type"] == "http.response.body":
                # TODO: check "more_body" and if true then do
                #  self.server.client_connection.send() to send the current data
                self.child.response.body.write(data["body"])
        else:
            raise AttributeError


class WebSocketHandler(Handler):
    def ws_shake_hand(self) -> None:
        pass

    def create_websocket(self) -> None:
        pass

    def send(self, data: Message) -> None:
        pass


class LifespanHandler(Handler):
    @staticmethod
    def create_scope() -> Scope:
        return {"type": "lifespan", "asgi": ASGI_VERSION}

    async def receive(self) -> Message:
        return {
            "type": "lifespan.startup"
            if not self.server.shutting_down
            else "lifespan.shutdown",
            "asgi": ASGI_VERSION,
        }

    async def send(self, data: Message) -> None:
        if (
            data["type"] == "lifespan.startup.failed"
            or data["type"] == "lifespan.shutdown.failed"
        ):
            if "startup" in data["type"]:
                self.server.server_log.error("App startup failed")
            if "shutdown" in data["type"]:
                self.server.server_log.error("App shutdown failed")
            self.server.server_log.error(data["message"])



from collections import Iterable as CollectionsInterable
from dataclasses import dataclass
from logging import Logger, getLogger
from typing import Any, Dict, Iterable, List, Union

from qactuar.models import Headers, Receive, Scope, Send


class BytesList:
    def __init__(self) -> None:
        self._bytes_list: List[bytes] = []

    def write(self, new_bytes: bytes) -> None:
        self._bytes_list.append(new_bytes)

    def writelines(self, *bytes_list: Union[bytes, Iterable[bytes]]) -> None:
        if len(bytes_list) == 1:
            if isinstance(bytes_list[0], CollectionsInterable):
                bytes_list = bytes_list[0]  # type: ignore
        self._bytes_list.extend(bytes_list)  # type: ignore

    def read(self) -> bytes:
        return b"".join(self._bytes_list)

    def readlines(self) -> List[bytes]:
        return self._bytes_list

    def clear(self) -> None:
        self._bytes_list = []

    def __contains__(self, item: bytes) -> bool:
        return item in self.read()

    def __len__(self) -> int:
        return len(self.read())

    def __bool__(self) -> bool:
        return len(self) > 0


def to_bytes(value: Any) -> bytes:
    if isinstance(value, bytes):
        return value
    return str(value).encode("utf-8")


@dataclass
class ProcStat:
    minor_faults: int = 0
    major_faults: int = 0
    user_time: int = 0
    system_time: int = 0
    virtual_size: int = 0
    resident_set_size: int = 0


@dataclass
class IoStat:
    read_char: int = 0
    write_char: int = 0
    read_bytes: int = 0
    write_bytes: int = 0


@dataclass
class MemStat:
    file_desc_size: int = 0
    vm_peak: int = 0
    vm_size: int = 0
    vm_highwm: int = 0
    vm_rss: int = 0
    vm_data: int = 0
    vm_lib: int = 0
    vm_pte: int = 0
    threads: int = 0


def parse_proc_stat(stat_line: str) -> ProcStat:
    parts = stat_line.split(" ")
    return ProcStat(
        minor_faults=int(parts[9]),
        major_faults=int(parts[11]),
        user_time=int(parts[13]),
        system_time=int(parts[14]),
        virtual_size=int(parts[22]),
        resident_set_size=int(parts[23]),
    )


io_stat_map = {
    "rchar": "read_char",
    "wchar": "write_char",
    "read_bytes": "read_bytes",
    "write_bytes": "write_bytes",
}


def parse_proc_io(io_line: str) -> IoStat:
    io_stat = IoStat()

    parts = io_line.split("\n")
    for part in parts:
        stat_parts = part.split(": ")
        if len(stat_parts) == 2:
            key, value = stat_parts
            if key in io_stat_map:
                setattr(io_stat, io_stat_map[key], int(value))

    return io_stat


mem_stat_map = {
    "FDSize": "file_desc_size",
    "VmPeak": "vm_peak",
    "VmSize": "vm_size",
    "VmHWM": "vm_highwm",
    "VmRSS": "vm_rss",
    "VmData": "vm_data",
    "VmLib": "vm_lib",
    "VmPTE": "vm_pte",
    "Threads": "threads",
}


def parse_proc_mem(mem_line: str) -> MemStat:
    mem_stat = MemStat()

    parts = mem_line.split("\n")
    for part in parts:
        stat_parts = part.split(":\t")
        if len(stat_parts) == 2:
            key, value = stat_parts
            if key in mem_stat_map:
                value = value.replace("kB", "")
                setattr(mem_stat, mem_stat_map[key], int(value))

    return mem_stat


try:
    import tornado  # type: ignore
except ImportError:
    pass
else:
    from typing import Type

    from tornado.http1connection import HTTP1Connection  # type: ignore
    from tornado.httputil import (  # type: ignore
        HTTPServerRequest,
        HTTPHeaders,
        RequestStartLine,
    )
    from tornado.iostream import BaseIOStream  # type: ignore
    from tornado.web import Application, RequestHandler, _HeaderTypes  # type: ignore

    # noinspection PyAbstractClass
    class QactuarStream(BaseIOStream):
        def __init__(self, *args: List[Any], **kwargs: Dict[Any, Any]) -> None:
            super().__init__(*args, **kwargs)

        def close_fd(self) -> None:
            pass

        def write_to_fd(self, data: memoryview) -> int:
            try:
                return len(data.tobytes())
            finally:
                del data

        def read_from_fd(self, buf: Union[bytearray, memoryview]) -> int:
            return len(buf)

        def fileno(self) -> int:
            return 0

    class TornadoWrapper:
        """
        Wraps a Tornado request handler object to provide a translation layer from the
        ASGI server to the handler. If anyone with more experience with Tornado could
        provide more insight/guidance with this wrapper and what it should be doing
        instead then just let me know.

        Right now it mocks some required objects and injects them where necessary to get
        it to run and then co-opts the write, add_header and set_header methods of the
        handler to get to the response data to send to the ASGI server.
        """

        def __init__(self, tornado_handler: Type[RequestHandler]) -> None:
            self.child_log = getLogger("qt_child")
            self.access_log = getLogger("qt_access")
            self.handler_type = tornado_handler

        async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
            # noinspection PyAbstractClass,PyMethodParameters
            class QactuarHandler(self.handler_type):  # type: ignore
                def __init__(
                    inner_self,
                    tornado_application: Application,
                    tornado_request: HTTPServerRequest,
                    logger: Logger,
                    **kwargs: Dict[Any, Any],
                ) -> None:
                    super().__init__(tornado_application, tornado_request, **kwargs)
                    inner_self._qactuar_body: List[bytes] = []
                    inner_self._qactuar_headers: Headers = []
                    inner_self._logger = logger

                def write(inner_self, chunk: Union[str, bytes, dict]) -> None:
                    super().write(chunk)
                    inner_self._qactuar_body.append(to_bytes(chunk))

                def add_header(inner_self, name: str, value: _HeaderTypes) -> None:
                    super().add_header(name, value)
                    inner_self._qactuar_headers.append(
                        (to_bytes(name), to_bytes(value))
                    )

                def set_header(inner_self, name: str, value: _HeaderTypes) -> None:
                    super().set_header(name, value)
                    inner_self._qactuar_headers.append(
                        (to_bytes(name), to_bytes(value))
                    )

            if scope["type"] == "http":
                received = await receive()
                body = received["body"]
                headers = HTTPHeaders(
                    {
                        header[0].decode("utf-8"): header[1].decode("utf-8")
                        for header in scope["headers"]
                    }
                )
                request_start_line = RequestStartLine(
                    scope["method"], scope["path"], scope["http_version"]
                )
                # noinspection PyTypeChecker
                http_connection = HTTP1Connection(QactuarStream(), False)
                http_connection._request_start_line = request_start_line
                http_connection._request_headers = headers
                request = HTTPServerRequest(
                    method=scope["method"],
                    uri=scope["path"],
                    version=scope["http_version"],
                    headers=headers,
                    body=body,
                    host=scope["server"][0],
                    connection=http_connection,
                    start_line=request_start_line,
                )
                handler = QactuarHandler(Application(), request, self.child_log)
                handler._transforms = []
                handler.application.transforms = []
                method_map = {
                    "GET": handler.get,
                    "POST": handler.post,
                    "PUT": handler.put,
                    "DELETE": handler.delete,
                    "OPTION": handler.options,
                    "HEAD": handler.head,
                    "PATCH": handler.patch,
                }
                try:
                    method = method_map[scope["method"]]
                    await method()
                except Exception as err:
                    self.access_log.exception(err)
                    result = b"500 Server Error"
                    handler.set_status(500)
                else:
                    result = b"".join(handler._qactuar_body)
                status = str(handler.get_status())
                response_headers = handler._qactuar_headers
                await send(
                    {
                        "type": "http.response.start",
                        "status": status,
                        "headers": response_headers,
                    }
                )
                await send(
                    {"type": "http.response.body", "body": result, "more_body": False}
                )



import asyncio
import multiprocessing
import socket
import sys
from logging import getLogger
from pathlib import Path
from platform import system
from time import time
from typing import TYPE_CHECKING
from uuid import uuid4

from qactuar.exceptions import HTTPError
from qactuar.request import Request
from qactuar.response import Response
from qactuar.util import BytesList, parse_proc_io, parse_proc_mem, parse_proc_stat

if TYPE_CHECKING:
    from qactuar import QactuarServer, ASGIApp


class ChildProcess:
    def __init__(self, server: "QactuarServer", client_socket: socket.socket):
        self.loop = asyncio.new_event_loop()
        self.server = server
        self.child_log = getLogger("qt_child")
        self.access_log = getLogger("qt_access")
        self.exception_log = getLogger("qt_exception")
        self.client_socket = client_socket
        self.response: Response = Response()
        self.raw_request_data: bytes = b""
        self.request_data: Request = Request()
        self.request_id: str = str(uuid4())
        client_socket.settimeout(server.config.RECV_TIMEOUT)

    @property
    def app(self) -> "ASGIApp":
        current_path = self.request_data.path
        for route, app in self.server.apps.items():
            if route == "/":
                if current_path == route:
                    return app
            elif current_path.startswith(route):
                self.request_data.path = current_path.replace(route, "")
                return app
        app = self.server.apps.get("/")  # type: ignore
        if app:
            return app
        raise HTTPError(404)

    def handle_one_request(self) -> None:
        self.server.http_handler.set_child(self)
        try:
            self.get_request_data()
            if not self.raw_request_data:
                self.close_socket()
                return
            if (
                self.request_data.headers["Connection"]
                and self.request_data.headers["Upgrade"]
            ):
                self.server.websocket_handler.ws_shake_hand()
            self.loop.run_until_complete(
                self.app(
                    self.server.http_handler.create_scope(),
                    self.server.http_handler.receive,
                    self.server.http_handler.send,
                )
            )
        except KeyboardInterrupt:
            sys.exit(0)
        except HTTPError as err:
            self.response.status = str(err.args[0]).encode("utf-8")
            self.response.body.write(str(err.args[0]).encode("utf-8"))
        except Exception as err:
            self.exception_log.exception(err)
            self.response.status = b"500"
            self.response.body.write(b"Internal Server Error")
        finally:
            if self.response:
                self.access_log.info(
                    "",
                    extra={
                        "host": self.server.client_info[0],
                        "port": self.server.client_info[1],
                        "request_id": self.request_id,
                        "method": self.request_data.method,
                        "http_version": self.request_data.request_version_num,
                        "path": self.request_data.original_path or "/",
                        "status": self.response.status.decode("utf-8"),
                    },
                )
            self.finish_response()

    def get_request_data(self) -> None:
        request_data = BytesList()
        request = Request()
        start = time()

        while True:
            try:
                request_data.write(
                    self.client_socket.recv(self.server.config.RECV_BYTES)
                )
            except socket.timeout:
                if not len(request_data):
                    if time() - start > self.server.config.REQUEST_TIMEOUT:
                        self.child_log.debug(
                            "no data received from request, timing out"
                        )
                        break
                request.raw_request = request_data.read()
                if request.headers_complete:
                    content_length = request.headers["content-length"]
                    if content_length is not None and request.method != "GET":
                        if len(request.body) == int(content_length):
                            break
                        else:
                            continue
                    break

        self.request_data = request
        self.raw_request_data = request_data.read()

    # TODO: http.disconnect when socket closes
    def finish_response(self) -> None:
        try:
            if self.response:
                self.response.add_header("x-request-id", self.request_id)
                self.client_socket.sendall(self.response.to_http())
        except OSError as err:
            self.exception_log.exception(err)
        self.get_proc_stats()
        self.close_socket()
        sys.exit(0)

    def close_socket(self) -> None:
        try:
            self.client_socket.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        self.client_socket.close()

    def get_proc_stats(self) -> None:
        # see https://linux.die.net/man/5/proc
        if system() == "Linux" and self.server.config.GATHER_PROC_STATS:
            try:
                pid = multiprocessing.current_process().pid
                pid_path = Path("/proc") / str(pid)

                with (pid_path / "stat").open() as stat_file:
                    stats = stat_file.read()
                    sts = parse_proc_stat(stats)
                    self.child_log.info(sts)

                with (pid_path / "io").open() as io_file:
                    io = io_file.read()
                    ios = parse_proc_io(io)
                    self.child_log.info(ios)

                with (pid_path / "status").open() as status_file:
                    status = status_file.read()
                    stas = parse_proc_mem(status)
                    self.child_log.info(stas)

            except Exception as err:
                self.exception_log.exception(err)


def make_child(server: "QactuarServer", client_socket: socket.socket) -> None:
    child = ChildProcess(server, client_socket)
    child.handle_one_request()



import asyncio
import errno
import multiprocessing
import os
import select
import socket
import sys
from importlib import import_module
from logging import Logger, getLogger, setLoggerClass
from logging.config import dictConfig
from time import time
from typing import Dict, Optional, Tuple

from qactuar.child_process import make_child
from qactuar.config import Config, config_init
from qactuar.handlers import HTTPHandler, LifespanHandler, WebSocketHandler
from qactuar.logs import QactuarLogger
from qactuar.models import ASGIApp, Receive, Scope, Send


class QactuarServer(object):
    address_family = socket.AF_INET
    socket_type = socket.SOCK_STREAM
    socket_level = socket.SOL_SOCKET
    socket_opt_name = socket.SO_REUSEADDR
    request_queue_size = 65536

    def __init__(
        self,
        host: str = None,
        port: int = None,
        app: ASGIApp = None,
        config: Config = None,
    ):
        if os.name != "nt":
            multiprocessing.set_start_method("fork")
            self.is_posix = False
        else:
            self.is_posix = True

        self.config: Config = config or config_init()

        setLoggerClass(QactuarLogger)
        dictConfig(self.config.LOGS)

        self.server_log: Logger = getLogger("qt_server")
        self.exception_log: Logger = getLogger("qt_exception")

        self.host: str = host or self.config.HOST
        self.port: int = port or self.config.PORT

        self.listen_socket: socket.socket = socket.socket(
            self.address_family, self.socket_type
        )
        self.listen_socket.setsockopt(self.socket_level, self.socket_opt_name, 1)
        self.listen_socket.bind((self.host, self.port))
        self.listen_socket.listen(self.request_queue_size)

        self.admin_socket: socket.socket = socket.socket(
            self.address_family, self.socket_type
        )
        self.admin_socket.setsockopt(self.socket_level, self.socket_opt_name, 1)
        self.admin_socket.bind((self.config.ADMIN_HOST, self.config.ADMIN_PORT))
        self.admin_socket.listen(self.request_queue_size)

        self.server_name: str = socket.getfqdn(self.host)
        self.server_port: int = self.port

        self.client_info: Tuple[str, int] = ("", 0)
        self.scheme: str = "http"
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self.processes: Dict[int, multiprocessing.Process] = {}
        self.shutting_down: bool = False
        self.time_last_cleaned_processes: float = time()
        self.lifespan_handler: LifespanHandler = LifespanHandler(self)
        self.http_handler: HTTPHandler = HTTPHandler(self)
        self.websocket_handler: WebSocketHandler = WebSocketHandler(self)
        self.apps: Dict[str, ASGIApp] = {"/": app} if app else {}
        for route, app_path in self.config.APPS.items():
            module_str, app_str = app_path.split(":")
            app_module = import_module(module_str)
            self.apps[route] = getattr(app_module, app_str)
        if self.apps:
            self.start_up()
            self.serve_forever()
        else:
            self.server_log.error("No apps found")
            self.shut_down()

    def add_app(self, application: ASGIApp, route: str = "/") -> None:
        self.apps[route] = application

    def start_up(self) -> None:
        self.send_to_all_apps(
            self.lifespan_handler.create_scope(),
            self.lifespan_handler.receive,
            self.lifespan_handler.send,
        )
        self.server_log.info(
            f"Qactuar: Serving {self.scheme.upper()} on {self.host}:{self.port}"
        )

    def shut_down(self) -> None:
        self.shutting_down = True
        self.server_log.info("Shutting Down")
        self.send_to_all_apps(
            self.lifespan_handler.create_scope(),
            self.lifespan_handler.receive,
            self.lifespan_handler.send,
        )
        sys.exit(0)

    def send_to_all_apps(self, scope: Scope, receive: Receive, send: Send) -> None:
        for app in self.apps.values():
            self.loop.run_until_complete(app(scope, receive, send))

    def serve_forever(self) -> None:
        try:
            while True:
                ready_to_read, _, _ = select.select(
                    [self.listen_socket], [], [], self.config.SELECT_SLEEP_TIME
                )
                if ready_to_read:
                    client_socket = self.accept_client_connection()
                    if client_socket:
                        self.fork(client_socket)
                self.check_processes()
        except KeyboardInterrupt:
            self.shut_down()
        except Exception as err:
            self.exception_log.exception(err)

    def accept_client_connection(self) -> Optional[socket.socket]:
        try:
            client_socket, self.client_info = self.listen_socket.accept()
        except IOError as err:
            if err.args[0] != errno.EINTR:
                raise
            return None  # for mypy
        else:
            return client_socket

    def fork(self, client_socket: socket.socket) -> None:
        process = multiprocessing.Process(target=make_child, args=(self, client_socket))
        process.daemon = True
        try:
            process.start()
        except AttributeError as err:
            self.exception_log.exception(err)
            self.server_log.warning(f"Could not start process {process.ident}")
        else:
            ident = process.ident
            if ident:
                self.processes[ident] = process

    def check_processes(self) -> None:
        current_time = time()
        last_time = self.time_last_cleaned_processes
        if current_time - last_time > self.config.CHECK_PROCESS_INTERVAL:
            self.time_last_cleaned_processes = current_time
            for ident, process in list(self.processes.items()):
                if not process.is_alive():
                    process.close()
                    del self.processes[ident]
